""" 
    Create a price handler. Used to do all price commands that run throughout the application.
    This price handler should use lists, as it will be faster to handle.
"""

from jamboree import Jamboree, DBHandler
from crayons import (blue, green, magenta, 
                        cyan, yellow, red)


# # TODO: Rethink design here and come back to this.

class TradeHandler(DBHandler):
    def __init__(self, limit=500):
        super().__init__()
        self.entity = "_trades"
        self.required = {
            "episode": str,
            "exchange": str,
            "live": bool
        }
        self._limit = 500
    

    """ 
        ---------------------------------------------------------------------------------
        ----------------------------------- Properties ----------------------------------
        ---------------------------------------------------------------------------------
    """

    @property
    def limit(self):
        return self._limit

    def limit(self, _limit):
        self._limit = _limit


    """ 
        ----------------------------------------------------------------------------------
        ------------------------------- Accessor Functions -------------------------------
        ----------------------------------------------------------------------------------
    """


    def save_trade(self, trade, user_id):
        """ Save trade in 2 ways to make it easy to access those trades later. """
        asset_alt = {}
        user_alt = {}
        exchange_alt = {}
        specified_trade_alt = {} # a way to get specific trade information.
        pass
    
    
    def latest_trade(self):
        return {}



    def latest_trade_by_user(self):
        return {}


    def latest_trade_by_asset(self):
        return {}


    

    

    
    """ 
        ----------------------------------------------------------------------------------
        ------------------------------- Counting Functions -------------------------------
        ----------------------------------------------------------------------------------
    """
    

    def trade_count(self, alt={}) -> int:
        """ Get the number of trades we have active in the given exchange. """
        count = self.count(alt=alt)
        return count
    
    def trade_exchange_count(self, exchange:str):
        alt = {"exchange": exchange}
        count = self.count(alt=alt)
        return count
    
    def trade_user_count(self):
        pass
#     def price_count(self) -> int:
#         count = self.count()
#         return count

#     """ 
#         --------------------------------------------------------------------------------
#         ------------------------------- Saving Functions -------------------------------
#         --------------------------------------------------------------------------------
#     """



#     def add_assets(self, assets:List[str]):
#         """ Add a list of assets to monitor. Used for tracking multiple assets inside of a portfolio. """
#         if len(assets) == 0:
#             return
#         asset_list = self.assets
#         for _ass in assets:
#             asset_list.append(_ass)
        

#         self.save_monitored_assets(asset_list)

#     def add_asset(self, asset_name:str):
#         if asset_name == "":
#             return

#         asset_list = self.assets
#         asset_list.append(asset_name)
#         self.save_monitored_assets(asset_list)

#     def save_monitored_assets(self, assets:list):
#         alt = {"detail": "assets"}
#         monitored = {
#             "assets": assets
#         }
#         self.save(monitored, alt=alt)

#     def save_single_bar(self, bar):
#         self.save(bar)

#     def save_bar_multi_bar(self, name:str, bars:List[Dict]):
#         alt = {"asset": name}
#         if len(bars) == 0:
#             return
        
#         query = copy.copy(self._query)
#         episode = query.get("episode")
#         exchange = query.get("exchange")
#         print(f"Saving bars for {name}-{episode}-{exchange}")
#         self.save_many(bars, alt=alt) # This will yield a flag, because why not?

#     """ 
#         --------------------------------------------------------------------------------
#         ---------------------------------- Misc ----------------------------------------
#         --------------------------------------------------------------------------------
#     """


#     def generate_price_bars(self, name, _type, _len, base_price):
#         price = generate_super_price(base_price, _type, _len)
#         ps = len(price)
#         vol_multiplier = random.uniform(2, 20)
#         open_multiplier = np.absolute(np.random.normal(1, 0.05, size=ps))
#         high_multiplier = np.absolute(np.random.normal(1.12, 0.05, size=ps))
#         low_multiplier =  np.absolute(np.random.normal(0.92, 0.05, size=ps))
#         noise = np.absolute(
#             np.random.normal(1, 0.2, size=ps)
#         )


#         volume = (noise * price) * vol_multiplier
#         _open = price * open_multiplier
#         high = high_multiplier * price
#         _low = low_multiplier * price

#         bars = []
#         current_time = maya.now()._epoch
#         hourly_multiplier = 3600
#         for _ in list(range(ps)).sort(reverse=True):
#             bar = {
#                 "open": float(_open[_]),
#                 "close": float(price[_]),
#                 "high": float(high[_]),
#                 "volume": float(volume[_]),
#                 "low": float(_low[_]),
#                 "time": float(current_time + (_*hourly_multiplier))
#             }
#             bars.append(bar)
#         # # Create a bunch of bars here
#         self.asset_bars[name] = bars




#     def generate(self, starting_min=50, starting_max=2000, is_varied=True, excluded_assets=[]):
#         """ Generate price bars for assets """
#         assets = self.assets
#         if len(assets) == 0:
#             return
        
#         for excluded in excluded_assets:
#             assets.remove(excluded)
        
#         # Generate bars
#         mid_point = ((starting_min + starting_max)/2)
#         std = (mid_point/4)

        
#         dask_tasks = []
#         for asset in assets:
#             base_price = random.normalvariate(mid_point, std)
#             # price_quote = f"Create prices for {asset}, {base_price}"
#             _type = random.choice(["GBM", "HESTON", "MERTON"])
#             dask_task = dask.delayed(self.generate_price_bars)(asset, _type, 4000, base_price)
#             dask_tasks.append(dask_task) 
#         dask.compute(*dask_tasks)
#         for asset in assets:
#             self.save_bar_multi_bar(asset, self.asset_bars[asset])
         
#     def _reset_assets(self):
#         # Set the first asset
#         count = self.asset_count()
#         if count == 0:
#             # Save that we have no new assets
#             self.save_monitored_assets([])



#     """ ---------------------------------------------------------------------------
#         ---------------------------- Live Data Handling---------------------------- 
#         ---------------------------------------------------------------------------
#     """
    
#     def save_live_price(self, data:dict):
#         """ Save the live price"""
#         alt = {}
#         pass


    
#     """ ---------------------------------------------------------------------------
#         ---------------------------- RL-Like Functions ---------------------------- 
#         ---------------------------------------------------------------------------
#     """

#     def step(self):
#         """ """
#         pass
    
#     def get_step(self):
#         """
#             # GET STEP
#             ---
#             Get all of the data for the latest step without pushing it forward.

#             Only works when live is set to false. Should only use it when live is set to false. 
#         """
#         pass

#     def reset(self):
#         """ No idea what goes here. But it should reset something."""
#         self._reset_assets()

# def main():
#     asset_list = ["BTC", "ATL", "TRX", "ETH", "BCH", "XRP", "LTC", "EOS", "ADA", "XMR", "LINK", "HT"]
#     price_handler = PriceHandler()
#     jam = Jamboree()
#     # This live variable locks certain commands, and allows us to query faster. 
#     price_handler['episode'] = uuid.uuid4().hex # set this to live if we're trying to use live prices.
#     price_handler['exchange'] = "binance"
#     price_handler['live'] = False
#     price_handler.event = jam
#     price_handler.add_assets(asset_list)
#     price_handler.reset()
    

#     price_handler.generate()
#     print(price_handler.assets)


# if __name__ == "__main__":
#     main()


import uuid
import pandas as pd
import maya
from jamboree import Jamboree, DBHandler
# from contextlib import ContextDecorator
from typing import List, Dict, Any
from crayons import yellow, blue, green, magenta
from loguru import logger
from link_jam_handle.data.pricing.simulated import SimulatedPriceHandler
from cytoolz.itertoolz import pluck
from mlfinlab.portfolio_optimization.mean_variance import MeanVarianceOptimisation

"""
Note: Will remove soon. Going through a concept.
"""




class AllocationHandler(DBHandler):
    """Abstract handler that we use to keep track of information.
    """

    def __init__(self, limit=500, spacing=5, unit="hours"):
        super().__init__()
        self.entity = "allocation"
        self.required = {
            "user_id": str,
            "exchange": str,
            "episode": str,
            "live": bool
        }
        self._balance = 0
        self._limit = limit
        self._spacing = spacing
        self._unit = unit
        self.mean_variance_opt = MeanVarianceOptimisation()
    
    
        
    
    # -------------------------------------------------------------------
    # --------------------- Properties & Setters ------------------------
    # -------------------------------------------------------------------
    


    @property
    def limit(self):
        return self._limit
    
    @limit.setter
    def limit(self, limit):
        self._limit = limit
    

    @property
    def spacing(self):
        return self._spacing

    @spacing.setter
    def spacing(self, space):
        self._spacing = space



    @property
    def unit(self):
        return self._unit
    
    @unit.setter
    def unit(self, _unit):
        self._unit = _unit


    @property
    def assets(self) -> list:
        """ Get all of the assets that we're watching for the exchange """
        return self.latest_asset().get("assets", [])

    @property
    def allocation(self):
        return self.latest_allocation().get("allocation", {})
    
    @property
    def current_step(self):
        alt = {"detail": "steps"}
        count = self.count(alt=alt)
        return count
    
    @property
    def last_time(self):
        _late = self.latest_timestamp()
        
        return self.latest_timestamp().get("epoch", maya.now()._epoch)

    @property
    def should_rebalance(self) -> bool:
        """ Depending on the scenario, we determine if we should rebalance."""
        """
            if live:
                check if the determined amount of time has passed since we've rebalanced
            else:
                check if the predetermined number of steps have happened.
        """
        if self["live"]:
            current_epoch = maya.now()._epoch
            expected = maya.MayaDT(self.last_time).add(**{f"{self.unit}":self.spacing})._epoch
            # print(expected, current_epoch)
            if current_epoch >= expected:
                
                return True
        else:
            if self.current_step % self.spacing == 0:
                logger.info(green("Yes, we're going to rebalance", bold=True))
                return True
        return False

    def latest_asset(self):
        alt = {"detail": "assets"}
        asset = self.last(alt=alt)
        return asset
    
    def latest_allocation(self):
        alt = {"detail": "allocation"}
        asset = self.last(alt=alt)
        return asset
    

    def latest_timestamp(self):
        alt = {"detail": "timestamp"}
        asset = self.last(alt=alt)
        return asset


    def timestamp_count(self) -> int:
        """ Get the count of the asset adds we have. Not the number of assets. """
        alt = {"detail": "timestamp"}
        count = self.count(alt=alt)
        return count

    
    def asset_count(self) -> int:
        """ Get the count of the asset adds we have. Not the number of assets. """
        alt = {"detail": "assets"}
        count = self.count(alt=alt)
        return count
    


    def count_steps(self):
        alt = {"detail": "steps"}
        count = self.count(alt=alt)
        return count


    def add_assets(self, assets:List[str]):
        """ Add a list of assets to monitor. Used for tracking multiple assets inside of a portfolio. """
        if len(assets) == 0:
            return
        asset_list = self.assets
        for asset in assets:
            if asset not in asset_list:
                asset_list.append(asset)
        self.save_monitored_assets(asset_list)

    def add_asset(self, asset_name:str):
        """ Add an asset that will be monitored in an exchange"""
        if asset_name == "":
            return
        asset_list = self.assets
        asset_list.append(asset_name)
        self.save_monitored_assets(asset_list)
        
    
    def save_monitored_assets(self, assets:list):
        """ Save monitored assets """
        alt = {"detail": "assets"}
        monitored = {
            "assets": assets
        }
        self.save(monitored, alt=alt)
    
    def save_allocation(self, allocation:dict):
        """ Save monitored assets """
        alt = {"detail": "allocation"}
        monitored = {
            "allocation": allocation
        }
        self.save(monitored, alt=alt)
    
    def save_timestamp(self, _time=None):
        """ Save monitored assets """
        if _time is None:
            _time = maya.now()._epoch
        alt = {"detail": "timestamp"}
        monitored = {
            "epoch": _time
        }
        self.save(monitored, alt=alt)


    def save_step(self):
        alt = {"detail": "steps"}
        monitored = {}
        self.save(monitored, alt=alt)
    







    def _filter_assets(self, all_assets:dict) -> dict:
        assets = self.assets
        users_portfolio_to_rebalance = {}
        _all_assets = all_assets.get("assets", None)
        if _all_assets is None:
            return {}
        for asset in assets:
            current_asset = _all_assets.get(asset, None)
            if current_asset is not None:
                users_portfolio_to_rebalance[asset] = current_asset

        return users_portfolio_to_rebalance
    

    def _extract_frame(self, portfolio_assets:dict):
        time_is_filled = False
        time_index = []
        portfolio_close = {}
        for asset in portfolio_assets.keys():
            current_pricing = portfolio_assets.get(asset, [])
            portfolio_close[asset] = list(pluck("close", current_pricing))
            if time_is_filled == False:
                time_index = list(pluck("time", current_pricing))
                time_is_filled = True
        
        pandas_index = pd.to_datetime(time_index, unit='s')
        pclose = pd.DataFrame(portfolio_close, index=pandas_index)
        return pclose

    

    


    def _finish_step(self, is_rebalance=False):
        """ Finish update that we're not """
        if is_rebalance == True:
            # Give a timestamp of the last rebalance.
            print(yellow("saving timestamp"))
            self.save_timestamp()
        self.save_step()
        


    def step(self, all_assets:Dict[str, Any]) -> None:
        """ Accept all of the assets for a given allocation and begin"""
        

        if self.should_rebalance == True:
            users_portfolio_to_rebalance = self._filter_assets(all_assets)

            portfolio_keys = users_portfolio_to_rebalance.keys()
            if len(portfolio_keys) == 0:
                logger.info(magenta("There are no portfolio keys", bold=True))
                self._finish_step()
                return

            pclose = self._extract_frame(users_portfolio_to_rebalance)
            self.mean_variance_opt.allocate(pclose)
            current_allocation = self.mean_variance_opt.weights.to_dict('records')[0]
            logger.info(current_allocation)
            self.save_allocation(current_allocation)
            self._finish_step(is_rebalance=True)
            return
            # Set timestamp aka, the last time we did a rebalance.
        self._finish_step()


    def _reset_assets(self) -> None:
        if self.asset_count() == 0:
            logger.info(yellow("Setting the first assets"))
            asset_list = self.assets
            self.save_monitored_assets(asset_list)
        else:
            logger.info(blue("Already reset"))

    def _reset_timestamp(self):
        if self.timestamp_count() == 0:
            logger.info("We're setting a null timestamp. This way the rebalance will happen automatically upon the first step.")
            self.save_timestamp(_time=maya.when('2 years ago')._epoch)


    def reset(self):
        self._reset_assets()
        self._reset_timestamp()


def experiment_main():
    jam = Jamboree()
    allo_handler = AllocationHandler(unit="seconds", spacing=2)
    price_handler = SimulatedPriceHandler()

    allo_handler.event = jam
    price_handler.event = jam
    
    
    
    allo_handler["user_id"] = "1"
    allo_handler["episode"] = uuid.uuid4().hex
    allo_handler["exchange"] = f"binance-{uuid.uuid4().hex}"
    allo_handler["live"] = True
    
    asset_list = ["BTC", "ATL", "TRX", "ETH", "BCH", "XRP", "LTC", "EOS", "ADA", "XMR", "LINK", "HT"]
    # This live variable locks certain commands, and allows us to query faster. 
    price_handler['episode'] = uuid.uuid4().hex # set this to live if we're trying to use live prices.
    price_handler['exchange'] = "binance"
    price_handler['live'] = False
    price_handler['obs_num'] = 2000
    price_handler.add_assets(asset_list)
    price_handler.reset()


    for _ in range(10):
        allo_handler.reset()
    allo_handler.add_assets(["BTC", "ETH", "TRX", "BCH"])


    while True:

        obs = price_handler.step()

        # Get the observation number
        i = obs.get("step_num", 0)
        assets = obs.get("assets", {})
        done = obs.get("done", False)
        if done:
            logger.info(blue("Observation Complete"))
            break
        # print()
        # latest_price = price_handler.latest_price("BTC")
        # logger.info(yellow(latest_price['time']))
        # logger.info(red(price_handler.current_time))

        allo_handler.step(obs)
    

if __name__ == "__main__":
    experiment_main()


import uuid
import maya
from jamboree import Jamboree, DBHandler
from contextlib import ContextDecorator


class timecontext(ContextDecorator):
    def __enter__(self):
        self.start = maya.now()._epoch
        return self

    def __exit__(self, *exc):
        self.end = maya.now()._epoch
        delta = self.end - self.start
        print(f"It took {delta}ms")
        return False

"""
Note: Will remove soon. Going through a concept.
"""

class UserHandler(DBHandler):
    """Abstract handler that we use to keep track of information.
    """

    def __init__(self, limit=100):
        # mongodb_host= "localhost", redis_host="localhost", redis_port=6379
        super().__init__()
        self.entity = "user"
        self.required = {
            "user_id": str
        }
        self._balance = 0
        self._limit = limit
        
    
    # -------------------------------------------------------------------
    # --------------------- Properties & Setters ------------------------
    # -------------------------------------------------------------------
    


    @property
    def limit(self):
        return self._limit
    
    @limit.setter
    def limit(self, limit):
        self._limit = limit
    

    def login(self, email, password):
        """ Check that the email and password """
        pass

    def register(self, username, password, confirm):
        pass

    
    def terms_and_conditions(self):
        pass

    def logout(self):
        pass
    


    # ------------------------------------------------------------------
    # -------------------- Queries I/O -------------------
    # ------------------------------------------------------------------

    def latest_user(self, email:str, episode:str="live"):
        """ Loads the latest user. Return None if they don't exist"""
        alt = {"detail": "user", "email":email, "episode": episode}
        pass

    # ----------------------------------------
    # -------------- Counting ----------------
    # ----------------------------------------

    # Use to get counts inside of the database




    def get_count(self) -> int:
        count = self.count()
        return count
    
    
    def count_user_records(self):
        """ Get the number of states that are available for the user """
        pass


    def reset(self):
        """ Determines if we're re-initiating """
        pass

# jambo = Jamboree()
# portfolio_handler = UserHandler()
# portfolio_handler.limit = 1000
# portfolio_handler.event = jambo
# portfolio_handler['episode'] = uuid.uuid1().hex
# portfolio_handler['user_id'] = uuid.uuid4().hex
# print(portfolio_handler.get_count())
# print(portfolio_handler.performance_count())

# portfolio_handler.reset()

# print(portfolio_handler.get_count())
# print(portfolio_handler.performance_count())
# print(portfolio_handler.balance)
# print(portfolio_handler.portfolio)


from .data.base_handlers import DataSet
from .data.base_handlers import RunnableSet
from .data.base_handlers import PriceSet
from .data.base_handlers.aggregate import AggregateDataset
from .data.base_handlers.live import LiveSet
from .data.base_handlers.backtest import BacktestDataset
from .allocation import AllocationHandler
from .exchange import PortfolioHandler
from .settings import SettingsHandler
from .trades import TradeHandler
from .user import UserHandler


from loguru import logger
import uuid
import copy
import maya
from jamboree import Jamboree, DBHandler
from contextlib import ContextDecorator
from link_jam_handle.utils.trades.trade_type import TradeType
from link_jam_handle.utils.trades.trade import Trade
from link_jam_handle import PriceSet
from link_jam_handle.data.pricing import SimulatedPriceHandler
from link_jam_handle.utils.slippage import RandomUniformSlippageModel
from cytoolz import keyfilter
from typing import Dict, List

def omit(blacklist, d):
    return keyfilter(lambda k: k not in blacklist, d)


class timecontext(ContextDecorator):
    def __enter__(self):
        self.start = maya.now()._epoch
        return self

    def __exit__(self, *exc):
        self.end = maya.now()._epoch
        delta = self.end - self.start
        print(f"It took {delta}ms")
        return False

"""
Note: Will remove soon. Going through a concept.
"""

class PortfolioHandler(DBHandler):
    """Abstract handler that we use to keep track of portfolio information.
    """

    def __init__(self, base_instrument="USD", start_balance=10000, limit=100, lazy_load=False):
        super().__init__()
        # TODO: Lazy load should be replaced with the parameters `pre_load` and `just_in_time`
        self._base_instrument = base_instrument
        self.start_balance = start_balance
        self.entity = "portfolio"
        self.required = {
            "episode": str,
            "user_id": str,
            "exchange": str
        }
        self._balance = 0
        self._limit = limit

        self.lazy = lazy_load
        # Store a list of the portfolio & performance states
        self._performance = []
        self._portfolio = []
        self._holdings = [] # Holdings is current_price * amount for each currency 
        self._trades = []
        self._pricing = None
        self._allocator = None
        self._slippage_model = RandomUniformSlippageModel()
        self._instrument_precision = 8
        self._commission_percent = 0.3
        self._base_precision = 2


        self._min_trade_amount = 0.000001
        self._max_trade_amount = 1000000
        self._min_trade_price = 0.00000001
        self._max_trade_price = 100000000



        # Placeholder variables
        self.placeholders = {}

        
        
        
    # -------------------------------------------------------------------
    # --------------------- Properties & Setters ------------------------
    # -------------------------------------------------------------------
    

    @property
    def current_time(self):
        """ 
            # Current Time
            ---
            Get the current time for the portfolio.

            Gets the time from the price/dataset. If the datasource is live it should get the current price and use that. 
        """
        return self._pricing.current_time


    @property
    def limit(self):
        return self._limit
    
    @limit.setter
    def limit(self, limit):
        self._limit = limit


    @property
    def holdings(self):
        """ Get the holdings for the user """
        _holdings = self.load_holdings()
        filtration_list = ['user_id', 'episode', 'type', 'detail', 'exchange', 'live', 'timestamp']
        holdings = map(lambda x: omit(filtration_list, x), _holdings)
        return list(holdings)


    @property
    def performance(self):
        """ Get the performance of the user/exchange"""
        performance = self.load_performance()
        if isinstance(performance, dict):
            performance = [performance]
        
        filtration_list = ["episode", "live", 'user_id', 'type', 'timestamp', 'exchange', 'detail']
        
        performance_filtered = list(map(lambda x: omit(filtration_list, x), performance))
        logger.info(performance_filtered)
        return performance_filtered


    @property
    def portfolio(self):
        """ Get the latest portfolio of the user/exchange"""
        if self.lazy == True:
            portfolio = self.latest_portfolio()
            filtered_portfolio = omit(['user_id', 'episode', 'type', 'timestamp'], portfolio) 
            return filtered_portfolio
        else:
            if isinstance(self._portfolio, dict):
                filtered_portfolio = omit(['user_id', 'episode', 'type', 'timestamp'], self._portfolio) 
                return filtered_portfolio
            filtered_portfolio = omit(['user_id', 'episode', 'type', 'timestamp'], self._portfolio[-1])
            return filtered_portfolio
            

    @property
    def balance(self) -> float:
        if self.lazy == True:
            latest_balance = self.latest_performance()
            return latest_balance.get('balance', 0.0)
        latest_record = self._performance[-1]
        balance = latest_record.get("balance", 0.0)
        return balance
    

    @property
    def trades(self) -> List[Dict]:
        """ Get all of the trades for this user"""
        trades = self.load_trades()
        if isinstance(trades, dict):
            trades = [trades]
        filtration_list = ["episode", "exchange", "user_id", "type", "detail", "timestamp", "live"]
        trades_filtered = map(lambda x: omit(filtration_list, x), trades)
        return list(trades_filtered)

    
    @property
    def net_worth(self) -> float:
        net_worth = self.balance
        portfolio = self.portfolio
        if not portfolio:
            return net_worth


        portfolio_filtered = omit(['user_id', 'episode', 'type', 'time', 'timestamp', 'exchange', 'live'], portfolio)

        for symbol, amount in portfolio_filtered.items():
            if symbol == self._base_instrument:
                continue

            current_price = self.current_price(symbol=symbol)
            net_worth += current_price * amount

        return net_worth

    @property
    def pricing(self) -> PriceSet:
        if self._pricing is None:
            raise NotImplementedError("You need to add pricing")
        return self._pricing
    
    @pricing.setter
    def pricing(self, price_handler: PriceSet):
        self._pricing = price_handler

    @property
    def allocator(self):
        """ Allocation stuff for"""
        if self._allocator is None:
            raise NotImplementedError("You need to add pricing")
        return self._allocator

    @allocator.setter
    def allocator(self, allocation_handler):
        self._allocator = allocation_handler

    @property
    def profit_loss_percent(self) -> float:
        """Calculate the percentage change in net worth since the last reset.
        Returns:
            The percentage change in net worth since the last reset.
        """
        return float(self.net_worth / self.start_balance) * 100
    # Use to get counts inside of the database


    def get_count(self) -> int:
        count = self.count()
        return count


    def performance_count(self) -> int:
        alt = {"detail": "performance"}
        count = self.count(alt)
        return count

    
    def holdings_count(self) -> int:
        alt = {"detail": "holdings"}
        count = self.count(alt)
        return count


    # ----------------------------------------
    # -------------- Querying ----------------
    # ----------------------------------------


    def latest_portfolio(self):
        last_state = self.last()
        return last_state

    def latest_performance(self):
        alt = {"detail": "performance"}
        last_state = self.last(alt)
        return last_state


    def load_performance(self):
        alt = {"detail": "performance"}
        performance = self.many(self.limit, alt=alt)
        if isinstance(performance, dict):
            return [performance]
        return performance
    
    
    def load_portfolio(self):
        portfolio = self.many(self.limit)
        if isinstance(portfolio, dict):
            return [portfolio]
        return portfolio
    

    def load_trades(self):
        alt = {"detail": "trade"}
        trades = self.many(self.limit, alt=alt)
        if isinstance(trades, dict):
            return [trades]
        return trades
    

    def load_holdings(self):
        alt = {"detail": "holdings"}
        _holdings = self.many(self.limit, alt=alt)
        if isinstance(_holdings, dict):
            return [_holdings]
        return _holdings


    # ----------------------------------------
    # ---------------- Saving ----------------
    # ----------------------------------------

    def save_portfolio(self, data):
        """ Save portfolio """
        _data = copy.copy(self._query)
        _data.update(data)
        _data['type'] = self.entity
        _data['time'] = self.current_time
        _data['timestamp'] = maya.now()._epoch

        self._portfolio.append(_data)
        self.save(_data)

    def save_performance(self, data) -> None:
        """ Save performance information """
        alt = {"detail": "performance"}
        _data = copy.copy(self._query)
        _data.update(data)
        _data.update(alt)
        _data['time'] = self.current_time
        _data['type'] = self.entity
        _data['timestamp'] = maya.now()._epoch
        
        self._performance.append(_data)
        self.save(_data, alt=alt)

    

    def save_trade(self, data):
        alt = {"detail": "trade"}

        _data = copy.copy(self._query)
        _data.update(data)
        _data.update(alt)

        _data['type'] = self.entity
        _data['time'] = self.current_time
        _data['timestamp'] = maya.now()._epoch
        self._trades.append(_data)
        logger.debug(_data)
        self.save(_data, alt=alt)


    def save_holdings(self, data, timestamp=maya.now()._epoch):
        alt = {"detail": "holdings"}

        _data = copy.copy(self._query)
        _data.update(data)
        _data.update(alt)

        _data['type'] = self.entity
        _data['time'] = self.current_time
        _data['timestamp'] = maya.now()._epoch
        self._trades.append(_data)
        self.save(_data, alt=alt)

    # ----------------------------------------
    # ------------ Operations ----------------
    # ----------------------------------------    

    
    def _update_account(self, _trade:Trade):
        """ Updates the portfolio exchange account. """
        # TODO: Add timestamp to trade
        
        if self._is_valid_trade(_trade):
            self._make_trade(_trade)

        self.placeholders['portfolio'][self._base_instrument] = self.placeholders['balance']


        latest_performance = {
            'net_worth': self.net_worth,
            'balance': self.placeholders['balance']
        }

        self.save_performance(latest_performance)
        self.save_portfolio(self.placeholders['portfolio'])
        self._update_holdings()
    
    def _update_account_no_order(self):
        """ Updates the portfolio exchange account. """
        # TODO: Add timestamp to trade

        self.placeholders['portfolio'][self._base_instrument] = self.placeholders['balance']


        latest_performance = {
            'net_worth': self.net_worth,
            'balance': self.placeholders['balance']
        }

        self.save_performance(latest_performance)
        self.save_portfolio(self.placeholders['portfolio'])
        self._update_holdings()


    def _update_holdings(self):
        """ Duplication of net_worth code"""
        holdings = {}
        portfolio = self.portfolio
        if not portfolio:
            return 


        portfolio_filtered = omit(['user_id', 'episode', 'type', 'time', 'timestamp', 'exchange', 'live'], portfolio)

        """
                "exchange": "superhotfyre",
                "episode": "335f32a64a4a4812b8373a65782e7801",
                "user_id": "10",
                "type": "portfolio",
                "detail": "trade",
                "live": false,
                "symbol": "ABC",
                "trade_type": 1,
                "amount": 1,
                "price": 792.64,
                "time": 1579585984.351019,
                "timestamp": 1577778784.3788486

        """


        portfolio_items = portfolio_filtered.items()
        for symbol, amount in portfolio_items:
            if symbol == self._base_instrument:
                holdings[symbol] = amount
                continue

            current_price = self.current_price(symbol=symbol)
            total = current_price * amount
            holdings[symbol] = total

        self.save_holdings(holdings, timestamp=self.current_time)

    def _make_trade(self, trade:Trade):
        """ Trades on the account then updates it"""

        if not trade.is_hold:
            """ TODO: We'll need to save this in two places to make it work IRL"""
            self.save_trade({
                'symbol': trade.symbol,
                'trade_type': trade.trade_type.value,
                'amount': trade.amount,
                'price': trade.price
            })
        
        if trade.is_buy:
            self.placeholders['balance'] -= trade.amount * trade.price
            self.placeholders['portfolio'][trade.symbol] = self.placeholders['portfolio'].get(trade.symbol, 0) + trade.amount
        elif trade.is_sell:
            self.placeholders['balance'] += trade.amount * trade.price
            self.placeholders['portfolio'][trade.symbol] = self.placeholders['portfolio'].get(trade.symbol, 0) - trade.amount
        
        


    def _is_valid_trade(self, _trade:Trade) -> bool:
        
        if _trade.is_buy and self.placeholders['balance'] < _trade.amount * _trade.price:
            logger.error("Not enough to buy")
            return False
        elif _trade.is_sell and self.placeholders['portfolio'].get(_trade.symbol, 0) < _trade.amount:
            logger.error("Not enough to sell")
            return False
        

        is_both = _trade.amount >= self._min_trade_amount and _trade.amount <= self._max_trade_amount

        if is_both: logger.success("Both conditions succeeded")
        return is_both


    def execute_trade(self, trade:Trade) -> Trade:
        self.placeholders['portfolio'] = self.latest_portfolio()
        self.placeholders['balance'] = self.balance

        current_price = self.current_price(symbol=trade.symbol)
        commission = self._commission_percent / 100
        filled_trade = trade.copy()
        

        if filled_trade.is_hold or not self._is_valid_trade(filled_trade):
            filled_trade.amount = 0



        if filled_trade.is_buy:
            price_adjustment = (1 + commission)
            filled_trade.price = round(current_price * price_adjustment, self._base_precision)
            filled_trade.amount = round((filled_trade.price * filled_trade.amount) / filled_trade.price,
                                        self._instrument_precision)
        elif filled_trade.is_sell:
            price_adjustment = (1 - commission)
            filled_trade.price = round(current_price * price_adjustment, self._base_precision)
            filled_trade.amount = round(filled_trade.amount, self._instrument_precision)

        if not filled_trade.is_hold:
            filled_trade = self._slippage_model.fill_order(filled_trade, current_price)
        

        self._update_account(filled_trade)
        return filled_trade

    def instrument_balance(self, symbol: str):
        """ Get the balance for the instrument """
        portfolio = self.latest_portfolio()

        if symbol in portfolio.keys():
            return portfolio[symbol]
        return 0.0

    

    
    # -------------------------------------------------------
    # ------------------ Reset Conditions -------------------
    # -------------------------------------------------------
    



    def _reset_portfolio(self):
        count = self.get_count()
        if count == 0:
            self.save_portfolio({f"{self._base_instrument}": float(self.start_balance), "time": self.current_time})

    def _reset_performance(self):
        """ Reset the performance inside of a given exchange """
        count = self.performance_count()
        if count == 0:
            self.save_performance({"balance": float(self.start_balance), "net_worth": float(self.start_balance), "time": self.current_time})
    
    def _reset_holdings(self):
        """ Reset the holdings inside of a given exchange """
        count = self.holdings_count()
        if count == 0:
            self.save_holdings({f"{self._base_instrument}": float(self.start_balance), "time": self.current_time})

    def _reset_price(self):
        """ Resets the price. If it's a backtest or simulation, we'll pull information into place so we can backtest. """
        self.pricing.reset()
    
    def _reset_allocation(self):
        self.allocator.reset()


    def update(self):
        """ Update the exchange. """
        self._update_holdings()


    def reset(self, _time=maya.now()._epoch):
        """ Determines if we're re-initiating """
        self._reset_price()
        self._reset_allocation()
        self._reset_performance()
        self._reset_portfolio()
        self._reset_performance()
        if self.lazy == False:
            """ This means we're loading all of the records upfront"""
            self._performance = self.load_performance()
            self._portfolio   = self.load_portfolio()
    

    def step(self, trade:Trade=None) -> dict:
        self.placeholders['portfolio'] = self.latest_portfolio()
        self.placeholders['balance'] = self.balance

        final_dict = {'obs': {}, 'done': False}
        prices_information = self.pricing.step()
        self.allocator.step(prices_information)
        if trade is not None:
            self.execute_trade(trade)
        final_dict['done'] = prices_information.get("done", True)
        final_dict['obs']["assets"] = prices_information.get("assets", {})
        # we'd put all of the portfolio information here by asset. 
        final_dict['obs']["portfolio"] = {}
        # get percentages
        return final_dict
    

    def step_no_trade(self) -> dict:
        self.placeholders['portfolio'] = self.latest_portfolio()
        self.placeholders['balance'] = self.balance

        final_dict = {'obs': {}, 'done': False}
        prices_information = self.pricing.step()
        self.allocator.step(prices_information)
        self._update_account_no_order()
        final_dict['done'] = prices_information.get("done", True)
        final_dict['obs']["assets"] = prices_information.get("assets", {})
        # we'd put all of the portfolio information here by asset. 
        final_dict['obs']["portfolio"] = {}
        # get percentages
        return final_dict

    def close(self):
        pass



def main():
    """ Should run through everything returning zero right now"""
    asset_list = ["BTC", "ATL", "TRX", "ETH", "BCH", "XRP", "LTC", "EOS", "ADA", "XMR", "LINK", "HT"]
    jambo = Jamboree()
    episode_id = uuid.uuid4().hex
    user_id = uuid.uuid4().hex
    portfolio_handler = PortfolioHandler()
    price_handler = SimulatedPriceHandler()
    price_handler['episode'] = episode_id # set this to live if we're trying to use live prices.
    price_handler['exchange'] = "binance"
    price_handler['live'] = False
    price_handler['obs_num'] = 1500
    price_handler.event = jambo
    price_handler.add_assets(asset_list)


    portfolio_handler.limit = 1000
    portfolio_handler.event = jambo
    portfolio_handler['episode'] = episode_id
    portfolio_handler['user_id'] = user_id
    portfolio_handler['exchange'] = "binance"
    portfolio_handler.pricing = price_handler
    portfolio_handler.reset()

    for _ in range(1000):
        step_information = portfolio_handler.step()
        logger.info(step_information.keys())




if __name__ == "__main__":
    main()


""" 
    Create a price handler. Used to do all price commands that run throughout the application.
    This price handler should use lists, as it will be faster to handle.
"""


from uuid import uuid1
from jamboree import Jamboree, DBHandler
from typing import Dict, Any, List


# NOTE: I'm probably going to have to separate everything here from

class SettingsHandler(DBHandler):
    """ Use this handler to manage the settings for things like data collection. """
    def __init__(self, limit=500):
        super().__init__()
        self.entity = "settings"
        self.required = {
            "organization": str,
            "live": bool
        }
        self._limit = 500
    

    """ 
        ---------------------------------------------------------------------------------
        ----------------------------------- Properties ----------------------------------
        ---------------------------------------------------------------------------------
    """

    @property
    def limit(self):
        return self._limit

    @limit.setter
    def limit(self, _limit):
        self._limit = _limit


    @property
    def tiingo_key(self):
        latest_key = self.latest_key("tiingo")
        return latest_key.get("key", "")

    @tiingo_key.setter
    def tiingo_key(self, _tiingo_key:str):
        self.save_key_information("tiingo", _tiingo_key, "")


    @property
    def main_tiingo_assets(self):
        """ Get the assets we're monitoring """
        api_name = "tiingo"
        exchange = "main"
        market = "us_stock"
        latest = self.latest_monitored_assets(api_name, exchange, market)
        return latest.get("assets", [])
    
    @main_tiingo_assets.setter
    def main_tiingo_assets(self, _assets):
        api_name = "tiingo"
        exchange = "main"
        market = "us_stock"

        if isinstance(_assets, str):
            self.add_asset(api_name, exchange, market, _assets)
        elif isinstance(_assets, list):
            self.add_assets(api_name, exchange, market, _assets)

    """ 
        ----------------------------------------------------------------------------------
        ------------------------------- Accessor Functions -------------------------------
        ----------------------------------------------------------------------------------
    """
    
    

    def latest_key(self, name):
        alt = {"detail": "key_info", "name": name}
        last_state = self.last(alt)
        last_state = last_state or {"key": "", "secret": "", "info": {}}
        return last_state

    
    def latest_monitored_assets(self, api_name:str, exchange:str, market:str):
        alt = {"detail": "assets", "api": api_name, "exchange": exchange, "market": market}
        latest = self.last(alt=alt)
        return latest

    """ 
        ----------------------------------------------------------------------------------
        ------------------------------- Counting Functions -------------------------------
        ----------------------------------------------------------------------------------
    """
    

    def complex_asset_count(self, api_name:str, exchange:str, market:str):
        alt = {"detail": "assets", "api": api_name, "exchange": exchange, "market": market}
        count = self.count(alt=alt)
        return count
    

    """ 
        --------------------------------------------------------------------------------
        ------------------------------- Saving Functions -------------------------------
        --------------------------------------------------------------------------------
    """

    def save_key_information(self, name:str, key:str, secret:str, info:Dict[str, Any]={}):
        alt = {"detail": "key_info", "name": name}
        monitored = {
            "key": key,
            "secret": secret,
            "info": info
        }
        self.save(monitored, alt=alt)


    def add_assets(self, api_name:str, exchange:str, market:str, assets:List[str]):
        """ Add a list of assets to monitor. Used for tracking multiple assets inside of a portfolio. """
        if len(assets) == 0:
            return
        
        latest = self.latest_monitored_assets(api_name, exchange, market)
        asset_list = latest.get("assets", [])
        for asset in assets:
            if asset not in asset_list:
                asset_list.append(asset)
        self.save_monitored_assets(api_name, exchange, market, asset_list)


    def add_asset(self, api_name:str, exchange:str, market:str, asset_name:str):
        """ Add an asset that will be monitored in an exchange"""
        if asset_name == "":
            return
        latest = self.latest_monitored_assets(api_name, exchange, market)
        asset_list = latest.get("assets", [])
        asset_list.append(asset_name)
        asset_list = list(set(asset_list))
        self.save_monitored_assets(api_name, exchange, market, asset_list)


    def save_monitored_assets(self, api_name:str, exchange:str, market:str, assets:list):
        """ Save monitored assets """
        alt = {"detail": "assets", "api": api_name, "exchange": exchange, "market": market}
        monitored = {
            "assets": assets
        }
        self.save(monitored, alt=alt)

    def _reset_tiingo_main(self):
        api_name = "tiingo"
        exchange = "main"
        market = "us_stock"

        if self.complex_asset_count(api_name, exchange, market) == 0:
            self.add_assets(api_name, exchange, market, [])

    def reset(self):
        """ No idea what goes here. But it should reset something."""

def main():
    setting_handler = SettingsHandler()
    jam = Jamboree()
    setting_handler['organization'] = "linkkt"
    setting_handler['live'] = False
    setting_handler.event = jam
    setting_handler.reset()
    setting_handler.tiingo_key = uuid1().hex
    # setting_handler.main_tiingo_assets = ["AAPL", "MSFT"]
    print(setting_handler.tiingo_key)
    print(setting_handler.main_tiingo_assets)


if __name__ == "__main__":
    main()


import os
import sys
from shutil import rmtree
from setuptools import setup, Command


here = os.path.abspath(os.path.dirname(__file__))


class UploadCommand(Command):
    """Support setup.py publish."""

    description = "Build and publish the package."
    user_options = []

    @staticmethod
    def status(s):
        """Prints things in bold."""
        print("\033[1m{0}\033[0m".format(s))

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        try:
            self.status("Removing previous builds…")
            rmtree(os.path.join(here, "dist"))
        except FileNotFoundError:
            pass
        self.status("Building Source distribution…")
        os.system("{0} setup.py sdist bdist_wheel".format(sys.executable))
        self.status("Uploading the package to PyPi via Twine…")
        os.system("sudo twine upload dist/*")
        sys.exit()

def read(fname):
    return open(os.path.join(os.path.dirname(__file__), fname)).read()

setup(
    name = "link_jam_handle",
    version="0.1",
    author = "Kevin Hill",
    author_email = "kah.kevin.hill@gmail.com",
    description = ("A bunch of database management for linkkt software."),
    license = "BSD",
    keywords = "example documentation tutorial",
    url = "http://packages.python.org/an_example_pypi_project",
    packages=['link_jam_handle'],
    long_description=read('README.md'),
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Topic :: Utilities",
        "License :: OSI Approved :: BSD License",
    ],
    install_requires=[
        'jamboree', 'mlfinlab', 'cytoolz', 'loguru', 'maya', 'crayons'
    ],
    cmdclass={"upload": UploadCommand},
)